/**********************************************************************
* FILE:		midi2pic.cpp									*
* OVERVIEW:	MIDI file to PIC 2-channel converter				*
* COPYRIGHT:	 M.J.Hutchby 2006								*
* AUTHOR:		James Hutchby									*
* UPDATED:	26/05/2006									*
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <alloc.h>
#include <string.h>


typedef int BOOL;
typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef char huge* HPSTR;

typedef DWORD FOURCC;		// four-character code
typedef FOURCC CKID;		// chunk identifier
typedef DWORD CKSIZE;		// 32-bit unsigned size


typedef struct {			// header chunk
	CKID ckID;		    	// type identifier
	CKSIZE ckSize;			// chunk size
	WORD Format;			// file format
	WORD NumTracks;		// number of tracks
	WORD Division;			// ticks per quarter note
} HCK;

typedef HCK huge* HPHCK;

typedef struct {			// track chunk
	CKID ckID;		    	// type identifier
	CKSIZE ckSize;			// chunk size
} TCK;

typedef TCK huge* HPTCK;


// returns a four-character code
inline FOURCC fourcc(char c1, char c2, char c3, char c4)
{
	return ((((FOURCC) c1) & 0xff) << 24) +
		  ((((FOURCC) c2) & 0xff) << 16) +
		  ((((FOURCC) c3) & 0xff) <<  8) +
		  ((((FOURCC) c4) & 0xff) <<  0);
}
inline FOURCC fourcc(HPSTR p)
{
	return ((((FOURCC) p[0]) & 0xff) << 24) +
		  ((((FOURCC) p[1]) & 0xff) << 16) +
		  ((((FOURCC) p[2]) & 0xff) <<  8) +
		  ((((FOURCC) p[3]) & 0xff) <<  0);
}


// gets a 24-bit value
inline DWORD get24(HPSTR p)
{
	return ((((DWORD) p[0]) & 0xff) << 16) +
		  ((((DWORD) p[1]) & 0xff) <<  8) +
		  ((((DWORD) p[2]) & 0xff) <<  0);
}

// gets a 32-bit value
inline DWORD get32(HPSTR p)
{
	return ((((DWORD) p[0]) & 0xff) << 24) +
		  ((((DWORD) p[1]) & 0xff) << 16) +
		  ((((DWORD) p[2]) & 0xff) <<  8) +
		  ((((DWORD) p[3]) & 0xff) <<  0);
}

// puts a 32-bit value
inline void put32(HPSTR p, DWORD x)
{
	p[0] = (BYTE) (x >> 24);
	p[1] = (BYTE) (x >> 16);
	p[2] = (BYTE) (x >>  8);
	p[3] = (BYTE) (x >>  0);
}


// converts little endian to big endian format
inline WORD bigendian(WORD x)
{
	return ((x & 0xff) << 8) + ((x >> 8) & 0xff);
}
inline DWORD bigendian(DWORD x)
{
	return (((x >>  0) & 0xff) << 24) +
		  (((x >>  8) & 0xff) << 16) +
		  (((x >> 16) & 0xff) <<  8) +
		  (((x >> 24) & 0xff) <<  0);
}


// chunk IDs
#define MThd_CHUNK fourcc('M','T','h','d')
#define MTrk_CHUNK fourcc('M','T','r','k')

// default extensions
#define MID ".mid"
#define ASM ".asm"

#define TRUE 1
#define FALSE 0

const long IBUFFER = 0x40000L;    	// input buffer size
const long OBUFFER = 0x10000L;    	// output buffer size

HPSTR Notes[] = {"C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"};

char ofile[256], ifile[256];		// filenames
char SongName[256] = ""; 		// song name
char TrackName[256] = ""; 		// track name
WORD Track;					// track number
WORD Track1, Track2;			// selected tracks
BYTE Channel1, Channel2;			// selected channels
DWORD Ticks;					// absolute time in ticks
DWORD Tempo = 500000L; 			// tempo in microseconds per quarter note (120BPM)
WORD Format;					// MIDI file format
WORD NumTracks;				// number of tracks
WORD Division;					// ticks per quarter note
double Scale;					// ticks scaling factor
BOOL EOT;						// end-of-track flag


// adds a default extension to a filename if absent
void add_extension(HPSTR filename, HPSTR extension)
{
	HPSTR p = filename;
	while (*p != '\0') p++;
	HPSTR end = p;

	if (end == filename) return;

	while (--p >= filename) {
		if (*p == '\\') break;
		if (*p == '.') return;
	}

	while ((*end++ = *extension++) != '\0') ;
}


// returns a variable-length encoded value
DWORD get_var_len(HPSTR& p)
{
	DWORD n = 0;
	do {n = (n << 7) + (*p & 0x7f);} while (*p++ & 0x80);
	return n;
}


// copies a name string
void copy_name(HPSTR s, DWORD len, HPSTR d)
{
	for (; len > 0; len--) *d++ = *s++;
	*d = '\0';
}


// processes a meta-event
void do_meta(HPSTR& s)
{
	BYTE meta = *s++;
	DWORD len = get_var_len(s);

	switch (meta) {

	// track sequence number
	case 0x00:
		break;

	// text
	case 0x01:
		break;

	// copyright
	case 0x02:
		break;

	// sequence or track name
	case 0x03:
		if (Track == 0 || *SongName == '\0') copy_name(s, len, SongName);
		copy_name(s, len, TrackName);
		if (*TrackName != '\0') printf("Track #%d: \"%s\"\n", Track, TrackName);
		break;

	// instrument name
	case 0x04:
		break;

	// lyric
	case 0x05:
		break;

	// marker
	case 0x06:
		break;

	// cue point
	case 0x07:
		break;

	// program (patch) name
	case 0x08:
		break;

	// device (port) name
	case 0x09:
		break;

	// MIDI channel prefix
	case 0x20:
		break;

	// end of track
	case 0x2F:
		EOT = TRUE;
		break;

	// tempo
	case 0x51:
		Tempo = get24(s);
		break;

	// SMPTE offset
	case 0x54:
		break;

	// time signature
	case 0x58:
		break;

	// key signature
	case 0x59:
		break;

	// proprietary information
	case 0x7F:
		break;

	default:
		break;
	}

	s += len;
}


// processes an event
void do_event(HPSTR& s, HPSTR& d1, HPSTR& d2)
{
	BYTE event, channel, note, velocity;
	DWORD len;

	// get delta-time
	DWORD delta = get_var_len(s);
	Ticks += delta;

	if (*s & 0x80) event = *s++;
	channel = (event & 0x0f) + 1;

	switch ((event & 0xf0) == 0xf0 ? event : (event & 0xf0)) {

	// note off
	case 0x80:
		note = *s++; velocity = *s++;
		if (Track == Track1 && (Channel1 == 0 || channel == Channel1)) {
			put32(d1, Ticks); d1 += 4;
			*d1++ = 0;
			*d1++ = note;
		}
		if (Track == Track2 && (Channel2 == 0 || channel == Channel2)) {
			put32(d2, Ticks); d2 += 4;
			*d2++ = 0;
			*d2++ = note;
		}
		break;

	// note on
	case 0x90:
		note = *s++; velocity = *s++;
		if (Track == Track1 && (Channel1 == 0 || channel == Channel1)) {
			put32(d1, Ticks); d1 += 4;
			if (velocity != 0) *d1++ = 1; else *d1++ = 0;
			*d1++ = note;
		}
		if (Track == Track2 && (Channel2 == 0 || channel == Channel2)) {
			put32(d2, Ticks); d2 += 4;
			if (velocity != 0) *d2++ = 1; else *d2++ = 0;
			*d2++ = note;
		}
		break;

	// after-touch
	case 0xa0:
		note = *s++; velocity = *s++;
		break;

	// control change
	case 0xb0:
		s++; s++;
		break;

	// program (patch) change
	case 0xc0:
		s++;
		break;

	// channel after-touch
	case 0xd0:
		s++;
		break;

	// pitch wheel change
	case 0xe0:
		s++; s++;
		break;

	// SYSEX
	case 0xf0:
		len = get_var_len(s);
		s += len;
		break;

	// song position pointer
	case 0xf2:
		s++; s++;
		break;

	// song select
	case 0xf3:
		s++;
		break;

	// tune request
	case 0xf6:
		break;

	// SYSEX continuation
	case 0xf7:
		len = get_var_len(s);
		s += len;
		break;

	// timing clock
	case 0xf8:
		break;

	// start current sequence
	case 0xfa:
		break;

	// continue stopped sequence
	case 0xfb:
		break;

	// stop sequence
	case 0xfc:
		break;

	// active sensing
	case 0xfe:
		break;

	// meta-event
	case 0xff:
		do_meta(s);
		break;

	default:
		while (!(*s & 0x80)) s++;
		break;
	}
}


// processes a track
void do_track(HPSTR& s, HPSTR& d1, HPSTR& d2)
{
	// get chunk identifier
	FOURCC ckID = fourcc(s); s += 4;

	// get chunk length
	DWORD length = get32(s); s += 4;

	HPSTR end = s + length;

	// ignore if not MIDI track chunk
	if (ckID != MTrk_CHUNK) {
		s = end;
		return;
	}

	Ticks = 0;

	EOT = FALSE;

	while (s < end) {

		do_event(s, d1, d2);

		if (EOT) break;
	}

	s = end;
}


// main entry point
int main(void)
{
	char buf[16];
	HPSTR p, s, d1, d2;

	// print sign-on message
	printf("MIDI file to PIC 2-channel converter\n\n");

	// allocate input buffer
	HPSTR ibuffer = (HPSTR) farmalloc(IBUFFER);
	if (ibuffer == NULL) {
		printf("Out of memory\n");
		return EXIT_FAILURE;
	}

	// allocate output buffers
	HPSTR obuffer1 = (HPSTR) farmalloc(OBUFFER);
	if (obuffer1 == NULL) {
		printf("Out of memory\n");
		return EXIT_FAILURE;
	}
	HPSTR obuffer2 = (HPSTR) farmalloc(OBUFFER);
	if (obuffer2 == NULL) {
		printf("Out of memory\n");
		return EXIT_FAILURE;
	}

	// get the input filename
	printf("Input MIDI filename (DOS 8.3 format) ? ");
	gets(ifile);

	if (*ifile == '\0') return EXIT_SUCCESS;

	// add the default extension
	add_extension(ifile, MID);

	// open the input file
	FILE* ip = fopen(ifile, "rb");
	if (ip == NULL) {
		printf("Can't open file %s\n", ifile);
		return EXIT_FAILURE;
	}

	// load the file
	p = ibuffer;
	unsigned long ilength = 0;
	int c;
	while ((c = fgetc(ip)) != EOF) {
		*p++ = c;
		if (++ilength >= IBUFFER) {
			printf("Input buffer overflow\n");
			return EXIT_FAILURE;
		}
	}

	// close the input file
	if (fclose(ip) == -1) {
		printf("Can't close file %s\n", ifile);
		return EXIT_FAILURE;
	}

	// check MIDI header chunk
	HPHCK hcp = (HPHCK) ibuffer;
	if (bigendian(hcp->ckID) != MThd_CHUNK) {
		printf("Invalid MIDI file format\n");
		return EXIT_FAILURE;
	}

	// check chunk length
	if (bigendian(hcp->ckSize) < 6L) {
		printf("Invalid header chunk length\n");
		return EXIT_FAILURE;
	}

	printf("\n" "File format = %d\n", Format = bigendian(hcp->Format));
	printf("Number of tracks = %d\n", NumTracks = bigendian(hcp->NumTracks));
	printf("Division = %d\n\n", Division = bigendian(hcp->Division));

	if (Division & 0x8000) {
		printf("Can't handle SMPTE time format\n");
		return EXIT_FAILURE;
	}

	// get tracks and channels
	if (Format == 0) {

		Track1 = 0;
		printf("MIDI channel to map to PIC channel #1 (1 to 16) ? ");
		gets(buf);
		if (*buf == '\0') Channel1 = 1; else Channel1 = (BYTE) atoi(buf);
		Track2 = 0;
		printf("MIDI channel to map to PIC channel #2 (1 to 16) ? ");
		gets(buf);
		if (*buf == '\0') Channel2 = 2; else Channel2 = (BYTE) atoi(buf);

	} else if (Format == 1) {

		printf("MIDI track to map to PIC channel #1 (0 to %d) ? ", NumTracks - 1);
		gets(buf);
		if (*buf == '\0') Track1 = 1; else Track1 = (BYTE) atoi(buf);
		printf("MIDI channel to map to PIC channel #1 (1 to 16, or 0 for all) ? ");
		gets(buf);
		if (*buf == '\0') Channel1 = 0; else Channel1 = (BYTE) atoi(buf);
		printf("MIDI track to map to PIC channel #2 (0 to %d) ? ", NumTracks - 1);
		gets(buf);
		if (*buf == '\0') Track2 = 2; else Track2 = (BYTE) atoi(buf);
		printf("MIDI channel to map to PIC channel #2 (1 to 16, or 0 for all) ? ");
		gets(buf);
		if (*buf == '\0') Channel2 = 0; else Channel2 = (BYTE) atoi(buf);

	} else if (Format == 2) {

		printf("MIDI track to map to PIC channel #1 (0 to %d) ? ", NumTracks - 1);
		gets(buf);
		if (*buf == '\0') Track1 = 0; else Track1 = (BYTE) atoi(buf);
		printf("MIDI channel to map to PIC channel #1 (1 to 16, or 0 for all) ? ");
		gets(buf);
		if (*buf == '\0') Channel1 = 1; else Channel1 = (BYTE) atoi(buf);
		printf("MIDI track to map to PIC channel #2 (0 to %d) ? ", NumTracks - 1);
		gets(buf);
		if (*buf == '\0') Track2 = 0; else Track2 = (BYTE) atoi(buf);
		printf("MIDI channel to map to PIC channel #2 (1 to 16, or 0 for all) ? ");
		gets(buf);
		if (*buf == '\0') Channel2 = 2; else Channel2 = (BYTE) atoi(buf);
	}

	printf("\n");

	s = (HPSTR) (hcp + 1);

	Track = 0;

	put32(obuffer1, -1L);
	put32(obuffer2, -1L);

	while (s < ibuffer + ilength) {

		HPSTR d1 = obuffer1, d2 = obuffer2;

		do_track(s, d1, d2);

		// end marker
		if (d1 != obuffer1) put32(d1, -1L);
		if (d2 != obuffer2) put32(d2, -1L);

		Track++;
	}

	printf("\n" "Tempo = %ld\n\n", Tempo);

	// get the output filename
	printf("Output assembler filename (DOS 8.3 format) ? ");
	gets(ofile);

	// exit if null filename
	if (*ofile == '\0') return EXIT_SUCCESS;

	// add the default extension
	add_extension(ofile, ASM);

	// open the output file
	FILE* op = fopen(ofile, "wt");
	if (op == NULL) {
		printf("Can't open file %s\n", ofile);
		return EXIT_FAILURE;
	}

	if (fprintf(op,
	  "; ***************************************\n"
	  "; ** MIDI file converted to 2 channels **\n"
	  "; ***************************************\n\n") == EOF) {
		printf("Can't write file %s\n", ofile);
		exit(EXIT_FAILURE);
	}

	if (*SongName != '\0' && fprintf(op, "; \"%s\"\n\n", SongName) == EOF) {
		printf("Can't write file %s\n", ofile);
		exit(EXIT_FAILURE);
	}

	if (fprintf(op,
	  "; channel #1 <- MIDI track #%d channel #%d (0 = all)\n"
	  "; channel #2 <- MIDI track #%d channel #%d (0 = all)\n\n\n",
	  Track1, Channel1, Track2, Channel2) == EOF) {
		printf("Can't write file %s\n", ofile);
		exit(EXIT_FAILURE);
	}

	// determine ticks scaling factor
	Scale = (((double) Tempo) / Division) / 2000;

	d1 = obuffer1; d2 = obuffer2;

	Ticks = 0;

	DWORD ticks1, ticks1_ = -1L, ticks2, ticks2_ = -1L;

	BYTE status1, note1, status1_ = -1, note1_ = -1;
	BYTE status2, note2, status2_ = -1, note2_ = -1;

	while (TRUE) {

		#define eot(d) ((get32(d) & 0x80000000L) != 0)
		if (eot(d1) && eot(d2)) break;

		ticks1 = get32(d1);
		ticks2 = get32(d2);

		DWORD delta;

		BYTE channel, octave, key;

		if (ticks1 <= ticks2) {

			delta = ticks1 - Ticks;

			d1 += 4; status1 = *d1++; note1 = *d1++;

			// ignore note played at same time as previous note
			// (i.e. only first note of chord played)
			if (status1 == 1 && ticks1 == ticks1_ && status1_ == 1) continue;

			// if no gap between notes
			if (status1 == 0 && get32(d1) == 0 && d1[4] == 1) continue;

			// note off only if matches most recent note on
			if (status1 == 0 && note1 != note1_) continue;

			channel = 0; octave = note1 / 12; key = note1 % 12;

		} else {

			delta = ticks2 - Ticks;

			d2 += 4; status2 = *d2++; note2 = *d2++;

			// ignore note played at same time as previous note
			// (i.e. only first note of chord played)
			if (status2 == 1 && ticks2 == ticks2_ && status2_ == 1) continue;

			// if no gap between notes
			if (status2 == 0 && get32(d2) == 0 && d2[4] == 1) continue;

			// note off only if matches most recent note on
			if (status2 == 0 && note2 != note2_) continue;

			channel = 1; octave = note2 / 12; key = note2 % 12;
		}

		// ignore notes outside central 4 octave range
		if (octave < 3 || octave > 6) continue;

		delta = (DWORD) (delta * Scale); delta >>= 2;

		if (delta & 0x4000) {
			printf("Delta-time too long\n");
			return EXIT_FAILURE;
		}

		BYTE hi = (BYTE) ((delta >> 7) & 0x7f);
		BYTE lo = (BYTE) (delta & 0x7f);

		if (hi == 0) {

			if (fprintf(op, "\tdt h'%02x',      ", lo) == EOF) {
				printf("Can't write file %s\n", ofile);
				exit(EXIT_FAILURE);
			}

		} else {

			if (fprintf(op, "\tdt h'%02x',h'%02x',", lo | 0x80, hi) == EOF) {
				printf("Can't write file %s\n", ofile);
				exit(EXIT_FAILURE);
			}
		}

		if ((channel == 0 ? status1 : status2) == 1) {

			BYTE n = (channel << 7) | ((octave - 3) << 4) | (key + 1);
			if (fprintf(op, "h'%02x'\t; %d:%s%d\n", n,
			  channel + 1, Notes[key], octave) == EOF) {
				printf("Can't write file %s\n", ofile);
				exit(EXIT_FAILURE);
			}

		} else {

			if (fprintf(op, "h'00'\t; %d:-\n", channel + 1) == EOF) {
				printf("Can't write file %s\n", ofile);
				exit(EXIT_FAILURE);
			}
		}

		Ticks = (channel == 0 ? ticks1 : ticks2);

		if (channel == 0) ticks1_ = ticks1, status1_ = status1, note1_ = note1;
		else ticks2_ = ticks2, status2_ = status2, note2_ = note2;
	}

	// end marker
	if (fprintf(op, "\n\tdt h'ff',h'ff'\n") == EOF) {
		printf("Can't write file %s\n", ofile);
		exit(EXIT_FAILURE);
	}

	// close the output file
	if (fclose(op) == -1) {
		printf("Can't close file %s\n", ofile);
		return EXIT_FAILURE;
	}

	// free buffers
	farfree(ibuffer);
	farfree(obuffer1);
	farfree(obuffer2);

	return EXIT_SUCCESS;
}

